unit uSimpleList;

interface

uses
  Generics.Collections, Generics.Defaults, TypInfo;

type

  TSimpleList<T> = class;

  TOnCompare<T> = function(ALeft, ARight: T): integer of object;

  TSimpleComparer<T> = class(TComparer<T>)
  public
    function Compare(const Left, Right: T): integer; override;
  public
    OnCompare: TOnCompare<T>;
  end;

  TSimpleList<T> = class(TList<T>)
  private

    FCurIndexPos: integer;
    FActiveIndex: integer;

    function DoPopByIndex(Index: integer): T;
    function FindByIndex(Index: integer): T;

    procedure FreeAllItems;
    procedure SetCurIndexPos(const Value: integer);
    procedure InnerRemoveduplicates(i: integer);
    procedure SetActiveIndex(const Value: integer);
    function GetReverseItem(Index: integer): T;

  protected

    FNeedFreeItem: boolean;

    procedure FreeItem(Item: T); virtual; // ؽȷͷ
    function SameItem(Item1, Item2: T): boolean; virtual;
    function Compare(ALeft, ARight: T): integer; virtual; // 

  public

    FSimpleComparer: TSimpleComparer<T>;

    constructor Create;
    destructor Destroy; override;

    procedure Lock; // °Lockֵѧϰ
    procedure Unlock; //

    function PopFirst: T; // ͣͬ
    function PopLast: T;
    function PopByIndex(Index: integer): T;
    function LastItem: T;
    function ActiveItem: T;
    function GetRandomItem: T;
    function GetByNegaive(AIndex: integer): T;

    procedure MoveItemsToMe(L: TSimpleList<T>);
    procedure FindAndFreeItem(Item: T);
    procedure RemoveAndFreeItem(AIndex: integer);
    procedure RemoveAllItemsWithoutFree;
    procedure ClearAndFreeAllItems; // ղͷеItem
    procedure RemoveDuplicates;
    procedure Shuffle; // ˳

    function AddItemWithoutExist(Item: T): boolean;

    property CurIndexPos: integer read FCurIndexPos write SetCurIndexPos;
    property ActiveIndex: integer read FActiveIndex write SetActiveIndex;

    property ReverseItem[Index: integer]: T read GetReverseItem;

  end;

  //  Constructor Ҫ T ҪһûCreate,Ҳǹ
  TClassSimpleList<T: Class, Constructor> = class(TSimpleList<T>)
  protected
    procedure FreeItem(Item: T); override;
  public
    function AddNewOne: T; virtual; // TCreate д ;
    procedure CreateItems(ACount: integer);
  end;

implementation

uses
  RTTI;

procedure TSimpleList<T>.MoveItemsToMe(L: TSimpleList<T>);
var
  Item: T;
begin
  while L.Count > 0 do
  begin
    Item := L.PopLast;
    self.Add(Item)
  end;
end;

function TSimpleList<T>.ActiveItem: T;
begin
  result := FindByIndex(ActiveIndex)
end;

function TSimpleList<T>.AddItemWithoutExist(Item: T): boolean;
var
  Q: T;
  p1, p2: pinteger;
  n1, n2: integer;
begin
  result := true;
  for Q in self do
  begin

    p1 := @Q;
    p2 := @Item;
    n1 := integer(p1^);
    n2 := integer(p2^);

    if n1 = n2 then
    begin
      result := false;
      exit;
    end;
  end;
  Add(Item);
end;

procedure TSimpleList<T>.ClearAndFreeAllItems;
var
  i: integer;
begin
  FreeAllItems;
  for i := self.Count - 1 downto 0 do
  begin
    Delete(i);
  end;
  clear;
  // Capacity := 0;
end;

function TSimpleList<T>.Compare(ALeft, ARight: T): integer;
begin
  result := 0;
end;

constructor TSimpleList<T>.Create;
begin
  FSimpleComparer := TSimpleComparer<T>.Create;
  inherited Create(FSimpleComparer);

  FSimpleComparer.OnCompare := self.Compare;
  FNeedFreeItem := true;
  FCurIndexPos := -1;
end;

destructor TSimpleList<T>.Destroy;
begin
  FreeAllItems;
  // FSimpleComparer.Free;
  inherited;
end;

function TSimpleList<T>.DoPopByIndex(Index: integer): T;
begin
  if (index >= 0) and (index <= Count - 1) then
  begin
    result := items[index];
    Delete(index);
    exit;
  end;
  result := T(nil);
end;

procedure TSimpleList<T>.FindAndFreeItem(Item: T);
var
  i: integer;
begin
  i := self.IndexOf(Item);
  if i > -1 then
  begin
    FreeItem(Item);
    Delete(i);
  end;
end;

function TSimpleList<T>.FindByIndex(Index: integer): T;
begin
  if (index >= 0) and (index <= Count - 1) then
  begin
    result := items[index];
    exit;
  end;
  result := T(nil);
end;

procedure TSimpleList<T>.FreeAllItems;
var
  Item: T;
  i: integer;
begin
  if FNeedFreeItem then
  begin
    FCurIndexPos := -1;
    for i := self.Count - 1 downto 0 do
    begin
      Item := self[i];
      FreeItem(Item);
    end;
  end;
end;

procedure TSimpleList<T>.FreeItem(Item: T);
var
  p: pointer;
begin
  //  T  PMyRec =^TMyRec  TMyRec=record;
  // д
  // if GetTypeKind(T) = tkPointer then
  // begin
  // Dispose(Pointer(Pointer(@Item)^));
  // end;
  // дδԲʹá
  //  Item ָ룬ڼ̳е FreeItem д Dispose(Item);
  // new(Item);
  p := typeinfo(T);
  // и p   _finalize

  // asm
  // mov eax,p
  // mov ebx,integer(item)
  // call
  // end;

end;

function TSimpleList<T>.GetByNegaive(AIndex: integer): T;
var
  n: integer;
begin
  n := Count + AIndex;
  result := FindByIndex(n);
end;

function TSimpleList<T>.GetRandomItem: T;
begin
  result := T(nil);
  if Count > 0 then
  begin
    Randomize;
    result := self[Random(Count)];
  end;
end;

function TSimpleList<T>.GetReverseItem(Index: integer): T;
begin
  result := self[Count - index - 1];
end;

procedure TSimpleList<T>.InnerRemoveduplicates(i: integer);
var
  j, k: integer;
  Item1, Item2: T;
begin

  if i >= 1 then
  begin

    k := i;

    Item1 := self[i];

    for j := i - 1 downto 0 do
    begin
      Item2 := self[j];
      if SameItem(Item1, Item2) then
      begin
        FreeItem(Item2);
        Delete(j);
        dec(k);
      end;
    end;

    InnerRemoveduplicates(k - 1);

  end;

end;

function TSimpleList<T>.LastItem: T;
begin
  result := FindByIndex(Count - 1)
end;

procedure TSimpleList<T>.Lock;
begin
  system.TMonitor.Enter(self);
end;

procedure TSimpleList<T>.Unlock;
begin
  system.TMonitor.exit(self);
end;

function TSimpleList<T>.PopByIndex(Index: integer): T;
begin
  result := DoPopByIndex(index);
end;

function TSimpleList<T>.PopFirst: T;
begin
  result := DoPopByIndex(0);
end;

function TSimpleList<T>.PopLast: T;
begin
  result := DoPopByIndex(Count - 1);
end;

procedure TSimpleList<T>.RemoveAllItemsWithoutFree;
var
  i: integer;
begin
  for i := self.Count - 1 downto 0 do
  begin
    Delete(i);
  end;
end;

procedure TSimpleList<T>.RemoveAndFreeItem(AIndex: integer);
var
  Item: T;
begin
  Item := self[AIndex];
  Delete(AIndex);
  FreeItem(Item);
end;

procedure TSimpleList<T>.RemoveDuplicates;
begin
  InnerRemoveduplicates(Count - 1);
end;

function TSimpleList<T>.SameItem(Item1, Item2: T): boolean;
begin
  result := false;
end;

procedure TSimpleList<T>.SetActiveIndex(const Value: integer);
begin
  FActiveIndex := Value;
end;

procedure TSimpleList<T>.SetCurIndexPos(const Value: integer);
begin
  FCurIndexPos := Value;
end;


procedure TSimpleList<T>.Shuffle;
var
  i: integer;
  n: integer;
  nCount: integer;
begin
  nCount := self.Count;
  for i := 0 to nCount - 1 do
  begin
    n := Random(nCount);
    Exchange(0, n);
  end;
end;

{ TThreadClassList<T> }

function TClassSimpleList<T>.AddNewOne: T;
begin
  result := T.Create();
  Add(result);
end;

procedure TClassSimpleList<T>.CreateItems(ACount: integer);
var
  i, n: integer;
begin
  n := ACount - self.Count;
  for i := 0 to n - 1 do
  begin
    AddNewOne;
  end;
end;

procedure TClassSimpleList<T>.FreeItem(Item: T);
begin
  Item.Free;
end;

{ TSimpleComparer<T> }

function TSimpleComparer<T>.Compare(const Left, Right: T): integer;
begin
  if Assigned(OnCompare) then
    result := OnCompare(Left, Right);
end;

end.
